//
// JTypist v3.1 - Java typing tutor program for UNIX systems
// Copyright (C) 1999  Simon Baldwin (simonb@sco.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//
// KBLayout class - defines a keyboard layout for the given locale
//
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

/**
 * Defines a keyboard layout for the given locale.
 */
public abstract class KBLayout {

	// description and keymap table; we use a vector since the ordering
	// of entries is used in ordering the statistics, and we want to
	// preserve it
	private String		kb_id			= null;
	private Vector		keymap_table		= new Vector();

	// lookup tables of glyphs by character they represent, and by keycode
	// they are assigned to, further looked up by shiftstate
	private Hashtable	glyphs_by_char		= new Hashtable();
	private Hashtable	glyphs_by_keycode_by_shiftstate
							= new Hashtable();


	/*
	 * Accessor methods.
	 */
	public String	get_kb_id()	{ return ( kb_id ); }
	public void	set_kb_id( String kb_id ) {
		this.kb_id = kb_id;
	}
	public Enumeration
			get_elements()	{ return ( keymap_table.elements() ); }


	/*
	 * Method to return the correct lookup table for a shiftstate.
	 */
	private Hashtable get_table( KeyShiftState shiftstate ) {

		// see if we already have a table for this shiftstate
		Hashtable hashtable = (Hashtable)
			(glyphs_by_keycode_by_shiftstate.get( shiftstate ));
		if ( hashtable == null ) {

			// not seen this shiftstate before, so create a new
			// lookup table for it
			hashtable = new Hashtable();
			glyphs_by_keycode_by_shiftstate.put(
							shiftstate, hashtable );
		}

		// return the right table for this shiftstate
		return ( hashtable );
	}

	/**
	 * Method to add a keymap entry into the table.  This places the new
	 * glyph at the end of the current table.
	 */
	public void append( Glyph glyph ) {

		// add the entry to the main list
		keymap_table.addElement( glyph );

		// add the entry to the lookup tables
		glyphs_by_char.put( glyph.get_character(), glyph );
		Hashtable hashtable = get_table( glyph.get_shiftstate() );
		hashtable.put( glyph.get_keycode(), glyph );
	}

	/**
	 * Method to replace an existing keymap entry in the table with
	 * another, retaining the position in the list of the original.
	 * The existing entry is a glyph having the same character.  If
	 * none exists, the glyph is added to the table.
	 */
	public void replace( Glyph glyph ) {

		// get the character for this glyph
		Character glyph_character	= glyph.get_character();

		// now search for any existing glyphs that implement the same
		// keyboard character
		Glyph old_glyph = (Glyph)
					(glyphs_by_char.get( glyph_character ));
		if ( old_glyph != null ) {

			// found one, so replace it with the new one, at the
			// same point in the original list
			int index = keymap_table.indexOf( old_glyph );
			keymap_table.setElementAt( glyph, index );

			// update the lookup tables for the replaced glyph
			glyphs_by_char.put( glyph.get_character(), glyph );
			Hashtable hashtable = get_table(
						glyph.get_shiftstate() );
			hashtable.put( glyph.get_keycode(), glyph );

			// since we found and replaced the glyph, we can
			// stop here
			return;
		}

		// there was nothing to replace, so instead the new glyph
		// is simply appended to the list
		append( glyph );
	}

	/**
	 * Method to delete a glyph from the keymap table.
	 */
	public void delete( char character ) {

		// search for any existing glyphs that implement the character
		Glyph old_glyph = (Glyph)(glyphs_by_char.get(
						new Character( character )));
		if ( old_glyph != null ) {

			// found a match, so remove it from the table
			keymap_table.removeElement( old_glyph );

			// remove the matched glyph from the lookup tables too
			glyphs_by_char.remove( old_glyph.get_character() );
			Hashtable hashtable = get_table(
						old_glyph.get_shiftstate() );
			hashtable.remove( old_glyph.get_keycode() );
		}
	}


	/**
	 * Find the glyph that represents a given character.  This returns
	 * a glyph, or null if no matching glyph exists.
	 */
	public Glyph find_glyph( char character ) {

		// return a simple lookup
		return ( (Glyph)(glyphs_by_char.get(
						new Character( character ))) );
	}

	/**
	 * Find the glyph that represents a given keycode, with shift state.
	 * This returns a glyph, or null if no matching glyph exists.
	 */
	public Glyph find_glyph( KeyCode keycode, KeyShiftState shiftstate ) {

		// return a simple lookup
		Hashtable hashtable = get_table( shiftstate );
		return ( (Glyph)(hashtable.get( keycode )) );
	}


	/**
	 * Printout method.
	 */
	public String toString() {
		return ( getClass().getName() + "{"
				+ "kb_id=\"" + kb_id
				+ "\",keymap_table@" + keymap_table
				+ "}" );
	}
}
